/**
 * @template T
 * @implements Iterator<T>
 * @implements Iterable<T>
 */
export class Iter {
  _iterator

  /**
   * @param {Iterator<T>} iterator 
   */
  constructor(iterator) {
    this._iterator = iterator
  }

  /**
   * @template T
   * @param {any} value 
   * @returns {value is Iterable<T>}
   */
  static _isIterable(value) {
    return Object.hasOwn(value, Symbol.iterator)
      && typeof value[Symbol.iterator] === 'function'
  }

  /**
   * @template T
   * @param {T} value 
   */
  static from(value) {
    if (value instanceof Iter) {
      return value
    }

    if (Iter._isIterable(value)) {
      const iterator = value[Symbol.iterator]()

      if (iterator instanceof Iter) {
        return iterator
      }

      return new Iter(iterator)
    }

    throw new TypeError('object is not an iterator')
  }

  /**
   * @param {any} [value]
   */
  next(value) {
    return this._iterator.next(value)
  }

  /**
   * @param {any} [value]
   */
  return(value) {
    // @ts-ignore
    return this._iterator.return(value)
  }

  /**
   * @param {any} err
   */
  throw(err) {
    // @ts-ignore
    return this._iterator.throw(err)
  }

  /**
   * @param {number} limit
   */
  drop(limit) {
    return new DropIter(this, limit)
  }

  /**
   * @param {(value: T, index: number) => boolean} callbackFn
   */
  every(callbackFn) {
    let next = this.next()
    let index = 0
    let result = true

    while (!next.done) {
      if (!callbackFn(next.value, index)) {
        result = false
        break
      }

      next = this.next()
      index += 1
    }

    this.return()
    return result
  }

  /**
   * @param {(value: T, index: number) => boolean} callbackFn
   */
  filter(callbackFn) {
    return new FilterIter(this, callbackFn)
  }

  /**
   * @param {(value: T, index: number) => boolean} callbackFn
   */
  find(callbackFn) {
    let next = this.next()
    let index = 0

    while (!next.done) {
      if (callbackFn(next.value, index)) {
        this.return()
        return next.value
      }

      next = this.next()
      index += 1
    }
  }

  /**
   * @param {<U>(value: T, index: number) => U} callbackFn
   */
  flatMap(callbackFn) {
    return new FlatMapIter(this, callbackFn)
  }

  /**
   * @param {(value: T, index: number) => void} callbackFn
   */
  forEach(callbackFn) {
    let next = this.next()
    let index = 0

    while (!next.done) {
      callbackFn(next.value, index)
      next = this.next()
      index += 1
    }
  }

  /**
   * @param {<U>(value: T, index: number) => U} callbackFn
   */
  map(callbackFn) {
    return new MapIter(this, callbackFn)
  }

  /**
   * @template U
   * @param {(accumulator: U, value: T, index: number) => U} callbackFn
   * @param {U} init 
   */
  reduce(callbackFn, init) {
    let next = this.next()
    let index = 0
    let acc = init

    while (!next.done) {
      acc = callbackFn(acc, next.value, index)

      next = this.next()
      index += 1
    }

    this.return()
    return acc
  }

  /**
   * @param {(value: T, index: number) => boolean} callbackFn
   */
  some(callbackFn) {
    let next = this.next()
    let index = 0
    let result = false

    while (!next.done) {
      if (callbackFn(next.value, index)) {
        result = true
        break
      }

      next = this.next()
      index += 1
    }

    this.return()
    return result
  }

  /**
   * @param {number} limit
   */
  take(limit) {
    return new TakeIter(this, limit)
  }

  /*
   * @returns {T[]}
   */
  toArray() {
    /** @type {T[]} */
    const result = []

    for (const item of this) {
      result.push(item)
    }

    return result
  }

  *[Symbol.iterator]() {
    return this
  }
}

/**
 * @template T
 * @extends Iter<T>
 */
class DropIter extends Iter {
  _limit

  /**
   * @param {Iterator<T>} iterator
   * @param {number} limit
   */
  constructor(iterator, limit) {
    super(iterator)
    this._limit = limit
  }

  /**
   * @param {any} value
   */
  next(value) {
    for (let i = this._limit; i > 0; i--) {
      const next = super.next(value)

      if (next.done) {
        return next
      }
    }

    return super.next(value)
  }
}

/**
 * @template T
 * @extends Iter<T>
 */
class FilterIter extends Iter {
  _filter
  _index = 0

  /**
   * @param {Iterator<T>} iterator
   * @param {(value: T, index: number) => boolean} callbackFn
   */
  constructor(iterator, callbackFn) {
    super(iterator)
    this._filter = callbackFn
  }

  /**
   * @param {any} [value]
   * @returns {IteratorResult<T>}
   */
  next(value) {
    let next = super.next(value)

    while (!next.done && !this._filter(next.value, this._index)) {
      next = super.next(value)
      this._index += 1
    }

    return next
  }
}

/**
 * @template T
 * @extends Iter<T>
 */
class FlatMapIter extends Iter {
  _flatMap
  _index = 0

  /** @type {Iterator<T> | undefined} */
  _inner = undefined

  /**
   * @param {Iterator<T>} iterator
   * @param {<U>(value: T, index: number) => Iterator<U> | Iterable<U>} callbackFn 
   */
  constructor(iterator, callbackFn) {
    super(iterator)
    this._flatMap = callbackFn
  }

  /**
   * @param {any} value
   * @returns {IteratorResult<T>}
   */
  next(value) {
    if (this._inner) {
      const innerResult = this._inner.next(value)

      if (!innerResult.done) {
        this._index += 1
        return { value: innerResult.value, done: false }
      }

      this._inner = undefined
    }

    const outerResult = super.next(value)
    if (outerResult.done) {
      return { value: undefined, done: true }
    }

    const nextIterable = this._flatMap(outerResult.value, this._index || 0)
    if (Iter._isIterable(nextIterable)) {
      this._inner = Iter.from(nextIterable)
      return this.next(value)
    } else {
      throw new TypeError('value is not an iterator')
    }
  }
}

/**
 * @template T
 * @extends Iter<T>
 */
class MapIter extends Iter {
  _map
  _index = 0

  /**
   * @param {Iterator<T>} iterator 
   * @param {<U>(value: T, index: number) => U} callbackFn 
   */
  constructor(iterator, callbackFn) {
    super(iterator)
    this._map = callbackFn
  }

  /** @param {any} value */
  next(value) {
    let next = super.next(value)

    if (next.done) {
      return next
    }

    const result = {
      done: false,
      value: this._map(next.value, this._index)
    }

    this._index += 1
    return result
  }
}


/**
 * @template T
 * @extends Iter<T>
 */
class TakeIter extends Iter {
  _limit

  /**
   * @param {Iterator<T>} iterator 
   * @param {number} limit
   */
  constructor(iterator, limit) {
    super(iterator)
    this._limit = limit
  }

  /**
   * @param {any} value
   * @returns {IteratorResult<T>}
   */
  next(value) {
    if (this._limit > 0) {
      const next = super.next(value)
      if (!next.done) {
        this._limit -= 1
      }
      return next
    }

    return { value: undefined, done: true }
  }
}

