import { diff, fail, fork, succeed, Tuple } from './fn.js'
import { ok } from '../vendor/kojima/src/index.js'
import { curry } from '../vendor/izuna/src/curry.js'
import { anyChar } from './char.js'

/** @import { ParseError, ParserState } from './state.js' */
/** @import { Result } from '../vendor/kojima/src/index.js' */

/**
 * @typedef {(value: any) => Result<ParserState, ParseError>} Parser
 */

export const take = n => state => {
  let result = anyChar(state)
  for (let i = n; i > 0; i--) {
    if (result.isErr()) {
      return result.chain(e => fail(`"take(${n})" failed`, state, e))
    }

    result = result.chain(anyChar)
  }

  return result
}

export const skip = parser => state => {
  const tokens = state[0]
  const result = parser(state)
  if (result.isOk()) {
    return result.map(other => [tokens, other[1]])
  } else {
    return result
  }
}

/**
 * @param {...Parser} parsers
 */
export const seq = (...parsers) =>
  /** @param {ParserState} state */
  state => {
    let acc = ok(state)

    for (const parser of parsers) {
      if (acc.isOk()) {
        acc = acc.bind(parser)
      } else {
        break
      }
    }

    return acc
  }

export const map = curry(
  /** 
   * @param {(...args: any[]) => any} fn
   * @param {Parser} parser
   * @param {ParserState} state
   */
  (fn, parser, state) => {
    return parser(state).map(result => {
      try {
        /** @type {Result<ParserState, ParseError>} */
        const parsed = result.chain(otherState =>
          fn(diff(state[0], otherState[0]))
        )

        const backtrack = result.chain(otherState =>
          Tuple(state[0], otherState[1])
        )

        return succeed(parsed, backtrack)
      } catch (e) {
        return fail('failed to map', state, e)
      }
    })
  })


export const many = curry((parser, state) => {
  let result = ok(state)

  while (true) {
    const [a, b] = fork(state)
    const res = parser(clone(result.unwrap()))
    if (res.isOk()) {
      result = res
    } else {
      break
    }
  }

  return result

})
